@@ -44,6 +44,13 @@ _SHEEPDOG_OPTS = [
     cfg.IntOpt('sheepdog_store_chunk_size',
                min=1,
                default=DEFAULT_CHUNKSIZE,
+               deprecated_for_removal=True,
+               deprecated_since='Train',
+               deprecated_reason="""
+The Sheepdog project is no longer actively maintained.  The
+Sheepdog driver is scheduled for removal in the 'U' development
+cycle.
+""",
                help="""
 Chunk size for images to be stored in Sheepdog data store.
 
@@ -67,6 +74,13 @@ Related Options:
 """),
     cfg.PortOpt('sheepdog_store_port',
                 default=DEFAULT_PORT,
+                deprecated_for_removal=True,
+                deprecated_since='Train',
+                deprecated_reason="""
+The Sheepdog project is no longer actively maintained.  The
+Sheepdog driver is scheduled for removal in the 'U' development
+cycle.
+""",
                 help="""
 Port number on which the sheep daemon will listen.
 
@@ -89,6 +103,13 @@ Related Options:
 """),
     cfg.HostAddressOpt('sheepdog_store_address',
                        default=DEFAULT_ADDR,
+                       deprecated_for_removal=True,
+                       deprecated_since='Train',
+                       deprecated_reason="""
+The Sheepdog project is no longer actively maintained.  The
+Sheepdog driver is scheduled for removal in the 'U' development
+cycle.
+""",
                        help="""
 Address to bind the Sheepdog daemon to.
 
@@ -122,6 +143,8 @@ class SheepdogImage(object):
         self.port = port
         self.name = name
         self.chunk_size = chunk_size
+        LOG.warning("The Sheepdog driver is deprecated and will be removed "
+                    "in the 'U' release.")
 
     def _run_command(self, command, data, *params):
         cmd = ['collie', 'vdi']
@@ -270,6 +293,10 @@ class Store(glance_store.driver.Store):
     def get_schemes(self):
         return ('sheepdog',)
 
+    def _set_url_prefix(self):
+        self._url_prefix = "%s://%s:%s:" % (
+            'sheepdog', self.addr, self.port)
+
     def configure_add(self):
         """
         Configure the Store to use the stored configuration options
@@ -304,6 +331,9 @@ class Store(glance_store.driver.Store):
             raise exceptions.BadStoreConfiguration(store_name='sheepdog',
                                                    reason=reason)
 
+        if self.backend_group:
+            self._set_url_prefix()
+
     @capabilities.check
     def get(self, location, offset=0, chunk_size=None, context=None):
         """
@@ -407,7 +437,7 @@ class Store(glance_store.driver.Store):
 
         metadata = {}
         if self.backend_group:
-            metadata['backend'] = u"%s" % self.backend_group
+            metadata['store'] = u"%s" % self.backend_group
 
         return (location.get_uri(),
                 offset,
