@@ -55,7 +55,7 @@ ONE_MB = units.k * units.Ki  # Here we used the mixed meaning of MB
 
 _SWIFT_OPTS = [
     cfg.BoolOpt('swift_store_auth_insecure', default=False,
-                help=_("""
+                help="""
 Set verification of the server certificate.
 
 This boolean determines whether or not to verify the server
@@ -70,10 +70,10 @@ Possible values:
 Related options:
     * swift_store_cacert
 
-""")),
+"""),
     cfg.StrOpt('swift_store_cacert',
                sample_default='/etc/ssl/certs/ca-certificates.crt',
-               help=_("""
+               help="""
 Path to the CA bundle file.
 
 This configuration option enables the operator to specify the path to
@@ -86,10 +86,10 @@ Possible values:
 Related options:
     * swift_store_auth_insecure
 
-""")),
+"""),
     cfg.StrOpt('swift_store_region',
                sample_default='RegionTwo',
-               help=_("""
+               help="""
 The region of Swift endpoint to use by Glance.
 
 Provide a string value representing a Swift region where Glance
@@ -115,13 +115,13 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.StrOpt('swift_store_endpoint',
                sample_default="""\
 https://swift.openstack.example.org/v1/path_not_including_container\
 _name\
 """,
-               help=_("""
+               help="""
 The URL endpoint to use for Swift backend storage.
 
 Provide a string value representing the URL endpoint to use for
@@ -140,10 +140,10 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.StrOpt('swift_store_endpoint_type', default='publicURL',
                choices=('publicURL', 'adminURL', 'internalURL'),
-               help=_("""
+               help="""
 Endpoint Type of Swift service.
 
 This string value indicates the endpoint type to use to fetch the
@@ -160,10 +160,10 @@ Possible values:
 Related options:
     * swift_store_endpoint
 
-""")),
+"""),
     cfg.StrOpt('swift_store_service_type',
                default='object-store',
-               help=_("""
+               help="""
 Type of Swift service to use.
 
 Provide a string value representing the service type to use for
@@ -181,10 +181,10 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.StrOpt('swift_store_container',
                default=DEFAULT_CONTAINER,
-               help=_("""
+               help="""
 Name of single container to store images/name prefix for multiple containers
 
 When a single container is being used to store images, this configuration
@@ -218,17 +218,17 @@ Related options:
     * ``swift_store_multi_tenant``
     * ``swift_store_create_container_on_put``
 
-""")),
+"""),
     cfg.IntOpt('swift_store_large_object_size',
                default=DEFAULT_LARGE_OBJECT_SIZE, min=1,
-               help=_("""
+               help="""
 The size threshold, in MB, after which Glance will start segmenting image data.
 
 Swift has an upper limit on the size of a single uploaded object. By default,
 this is 5GB. To upload objects bigger than this limit, objects are segmented
 into multiple smaller objects that are tied together with a manifest file.
 For more detail, refer to
-http://docs.openstack.org/developer/swift/overview_large_objects.html
+https://docs.openstack.org/swift/latest/overview_large_objects.html
 
 This configuration option specifies the size threshold over which the Swift
 driver will start segmenting image data into multiple smaller files.
@@ -244,10 +244,10 @@ Possible values:
 Related options:
     * ``swift_store_large_object_chunk_size``
 
-""")),
+"""),
     cfg.IntOpt('swift_store_large_object_chunk_size',
                default=DEFAULT_LARGE_OBJECT_CHUNK_SIZE, min=1,
-               help=_("""
+               help="""
 The maximum size, in MB, of the segments when image data is segmented.
 
 When image data is segmented to upload images that are larger than the limit
@@ -267,9 +267,9 @@ Possible values:
 Related options:
     * ``swift_store_large_object_size``
 
-""")),
+"""),
     cfg.BoolOpt('swift_store_create_container_on_put', default=False,
-                help=_("""
+                help="""
 Create container, if it doesn't already exist, when uploading image.
 
 At the time of uploading an image, if the corresponding container doesn't
@@ -284,9 +284,9 @@ Possible values:
 Related options:
     * None
 
-""")),
+"""),
     cfg.BoolOpt('swift_store_multi_tenant', default=False,
-                help=_("""
+                help="""
 Store images in tenant's Swift account.
 
 This enables multi-tenant storage mode which causes Glance images to be stored
@@ -305,10 +305,10 @@ Possible values:
 Related options:
     * swift_store_config_file
 
-""")),
+"""),
     cfg.IntOpt('swift_store_multiple_containers_seed',
                default=0, min=0, max=32,
-               help=_("""
+               help="""
 Seed indicating the number of containers to use for storing images.
 
 When using a single-tenant store, images can be stored in one or more than one
@@ -333,9 +333,9 @@ Related options:
     * ``swift_store_multi_tenant``
     * ``swift_store_create_container_on_put``
 
-""")),
+"""),
     cfg.ListOpt('swift_store_admin_tenants', default=[],
-                help=_("""
+                help="""
 List of tenants that will be granted admin access.
 
 This is a list of tenants that will be granted read/write access on
@@ -349,10 +349,10 @@ Possible values:
 Related options:
     * None
 
-""")),
+"""),
     cfg.BoolOpt('swift_store_ssl_compression',
                 default=True,
-                help=_("""
+                help="""
 SSL layer compression for HTTPS Swift requests.
 
 Provide a boolean value to determine whether or not to compress
@@ -373,11 +373,11 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.IntOpt('swift_store_retry_get_count',
                default=0,
                min=0,
-               help=_("""
+               help="""
 The number of times a Swift download will be retried before the
 request fails.
 
@@ -395,11 +395,11 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.IntOpt('swift_store_expire_soon_interval',
                min=0,
                default=60,
-               help=_("""
+               help="""
 Time in seconds defining the size of the window in which a new
 token may be requested before the current token is due to expire.
 
@@ -421,10 +421,10 @@ Possible values:
 Related Options:
     * None
 
-""")),
+"""),
     cfg.BoolOpt('swift_store_use_trusts',
                 default=True,
-                help=_("""
+                help="""
 Use trusts for multi-tenant Swift store.
 
 This option instructs the Swift store to create a trust for each
@@ -447,10 +447,10 @@ Possible values:
 Related options:
     * swift_store_multi_tenant
 
-""")),
+"""),
     cfg.BoolOpt('swift_buffer_on_upload',
                 default=False,
-                help=_("""
+                help="""
 Buffer image segments before upload to Swift.
 
 Provide a boolean value to indicate whether or not Glance should
@@ -473,7 +473,7 @@ Possible values:
 Related options:
     * swift_upload_buffer_dir
 
-"""))
+"""),
 ]
 
 
@@ -495,31 +495,34 @@ def swift_retry_iter(resp_iter, length, store, location, manager):
     retries = 0
     bytes_read = 0
 
-    while retries <= store.conf.glance_store.swift_store_retry_get_count:
+    if store.backend_group:
+        rcount = getattr(store.conf,
+                         store.backend_group).swift_store_retry_get_count
+    else:
+        rcount = store.conf.glance_store.swift_store_retry_get_count
+
+    while retries <= rcount:
         try:
             for chunk in resp_iter:
                 yield chunk
                 bytes_read += len(chunk)
         except swiftclient.ClientException as e:
-            LOG.warning(_("Swift exception raised %s")
-
+            LOG.warning("Swift exception raised %s"
                         % encodeutils.exception_to_unicode(e))
 
         if bytes_read != length:
-            if retries == store.conf.glance_store.swift_store_retry_get_count:
+            if retries == rcount:
                 # terminate silently and let higher level decide
                 LOG.error(_LE("Stopping Swift retries after %d "
                               "attempts") % retries)
                 break
             else:
                 retries += 1
-                glance_conf = store.conf.glance_store
-                retry_count = glance_conf.swift_store_retry_get_count
                 LOG.info(_LI("Retrying Swift connection "
                              "(%(retries)d/%(max_retries)d) with "
                              "range=%(start)d-%(end)d"),
                          {'retries': retries,
-                          'max_retries': retry_count,
+                          'max_retries': rcount,
                           'start': bytes_read,
                           'end': length})
                 (_resp_headers, resp_iter) = store._get_object(location,
@@ -578,7 +581,11 @@ class StoreLocation(location.StoreLocation):
         if not credentials_included:
             # Used only in case of an add
             # Get the current store from config
-            store = self.conf.glance_store.default_swift_reference
+            if self.backend_group:
+                store = getattr(self.conf,
+                                self.backend_group).default_swift_reference
+            else:
+                store = self.conf.glance_store.default_swift_reference
 
             return '%s://%s/%s/%s' % ('swift+config', store, container, obj)
         if self.scheme == 'swift+config':
@@ -593,7 +600,8 @@ class StoreLocation(location.StoreLocation):
 
     def _get_conf_value_from_account_ref(self, netloc):
         try:
-            ref_params = sutils.SwiftParams(self.conf).params
+            ref_params = sutils.SwiftParams(
+                self.conf, backend=self.backend_group).params
             self.user = ref_params[netloc]['user']
             self.key = ref_params[netloc]['key']
             netloc = ref_params[netloc]['auth_address']
@@ -685,8 +693,8 @@ class StoreLocation(location.StoreLocation):
             raise exceptions.BadStoreUri(message=reason)
 
         pieces = urllib.parse.urlparse(uri)
-        assert pieces.scheme in ('swift', 'swift+http', 'swift+https',
-                                 'swift+config')
+        self.validate_schemas(uri, valid_schemas=(
+            'swift://', 'swift+http://', 'swift+https://', 'swift+config://'))
 
         self.scheme = pieces.scheme
         netloc = pieces.netloc
@@ -726,11 +734,21 @@ class StoreLocation(location.StoreLocation):
             return ''.join([auth_scheme, self.auth_or_store_url])
 
 
-def Store(conf):
+def Store(conf, backend=None):
+    group = 'glance_store'
+    if backend:
+        group = backend
+        multi_tenant = getattr(conf, backend).swift_store_multi_tenant
+        default_store = conf.glance_store.default_backend
+    else:
+        default_store = conf.glance_store.default_store
+        multi_tenant = conf.glance_store.swift_store_multi_tenant
+
     # NOTE(dharinic): Multi-tenant store cannot work with swift config
-    if conf.glance_store.swift_store_multi_tenant:
-        if (conf.glance_store.default_store == 'swift+config' or
-           sutils.is_multiple_swift_store_accounts_enabled(conf)):
+    if multi_tenant:
+        if (default_store == 'swift+config' or
+                sutils.is_multiple_swift_store_accounts_enabled(
+                    conf, backend=backend)):
             msg = _("Swift multi-tenant store cannot be configured to "
                     "work with swift+config. The options "
                     "'swift_store_multi_tenant' and "
@@ -742,13 +760,13 @@ def Store(conf):
                                                    reason=msg)
     try:
         conf.register_opts(_SWIFT_OPTS + sutils.swift_opts +
-                           buffered.BUFFERING_OPTS, group='glance_store')
+                           buffered.BUFFERING_OPTS, group=group)
     except cfg.DuplicateOptError:
         pass
 
-    if conf.glance_store.swift_store_multi_tenant:
-        return MultiTenantStore(conf)
-    return SingleTenantStore(conf)
+    if multi_tenant:
+        return MultiTenantStore(conf, backend=backend)
+    return SingleTenantStore(conf, backend=backend)
 
 Store.OPTIONS = _SWIFT_OPTS + sutils.swift_opts + buffered.BUFFERING_OPTS
 
@@ -771,7 +789,11 @@ class BaseStore(driver.Store):
         return ('swift+https', 'swift', 'swift+http', 'swift+config')
 
     def configure(self, re_raise_bsc=False):
-        glance_conf = self.conf.glance_store
+        if self.backend_group:
+            glance_conf = getattr(self.conf, self.backend_group)
+        else:
+            glance_conf = self.conf.glance_store
+
         _obj_size = self._option_get('swift_store_large_object_size')
         self.large_object_size = _obj_size * ONE_MB
         _chunk_size = self._option_get('swift_store_large_object_chunk_size')
@@ -821,10 +843,14 @@ class BaseStore(driver.Store):
     @capabilities.check
     def get(self, location, connection=None,
             offset=0, chunk_size=None, context=None):
+        if self.backend_group:
+            glance_conf = getattr(self.conf, self.backend_group)
+        else:
+            glance_conf = self.conf.glance_store
+
         location = location.store_location
         # initialize manager to receive valid connections
-        allow_retry = \
-            self.conf.glance_store.swift_store_retry_get_count > 0
+        allow_retry = glance_conf.swift_store_retry_get_count > 0
         with self.get_manager(location, context,
                               allow_reauth=allow_retry) as manager:
             (resp_headers, resp_body) = self._get_object(location,
@@ -835,7 +861,7 @@ class BaseStore(driver.Store):
                     try:
                         return next(self.wrapped)
                     except StopIteration:
-                        return ''
+                        return b''
 
             length = int(resp_headers.get('content-length', 0))
             if allow_retry:
@@ -855,7 +881,11 @@ class BaseStore(driver.Store):
             return 0
 
     def _option_get(self, param):
-        result = getattr(self.conf.glance_store, param)
+        if self.backend_group:
+            result = getattr(getattr(self.conf, self.backend_group), param)
+        else:
+            result = getattr(self.conf.glance_store, param)
+
         if not result:
             reason = (_("Could not find %(param)s in configuration options.")
                       % param)
@@ -875,9 +905,28 @@ class BaseStore(driver.Store):
                 LOG.exception(msg % {'container': container,
                                      'chunk': chunk})
 
+    @driver.back_compat_add
     @capabilities.check
-    def add(self, image_id, image_file, image_size,
+    def add(self, image_id, image_file, image_size, hashing_algo,
             context=None, verifier=None):
+        """
+        Stores an image file with supplied identifier to the backend
+        storage system and returns a tuple containing information
+        about the stored image.
+
+        :param image_id: The opaque image identifier
+        :param image_file: The image data to write, as a file-like object
+        :param image_size: The size of the image data to write, in bytes
+        :param hashing_algo: A hashlib algorithm identifier (string)
+        :param verifier: An object used to verify signatures for images
+
+        :returns: tuple of URL in backing store, bytes written, checksum,
+                multihash value, and a dictionary with storage system
+                specific information
+        :raises: `glance_store.exceptions.Duplicate` if something already
+                exists at this location
+        """
+        os_hash_value = hashlib.new(str(hashing_algo))
         location = self.create_location(image_id, context=context)
         # initialize a manager with re-auth if image need to be splitted
         need_chunks = (image_size == 0) or (
@@ -894,17 +943,13 @@ class BaseStore(driver.Store):
                 if not need_chunks:
                     # Image size is known, and is less than large_object_size.
                     # Send to Swift with regular PUT.
-                    if verifier:
-                        checksum = hashlib.md5()
-                        reader = ChunkReader(image_file, checksum,
-                                             image_size, verifier)
-                        obj_etag = manager.get_connection().put_object(
-                            location.container, location.obj,
-                            reader, content_length=image_size)
-                    else:
-                        obj_etag = manager.get_connection().put_object(
-                            location.container, location.obj,
-                            image_file, content_length=image_size)
+                    checksum = hashlib.md5()
+                    reader = ChunkReader(image_file, checksum,
+                                         os_hash_value, image_size,
+                                         verifier=verifier)
+                    obj_etag = manager.get_connection().put_object(
+                        location.container, location.obj,
+                        reader, content_length=image_size)
                 else:
                     # Write the image into Swift in chunks.
                     chunk_id = 1
@@ -940,8 +985,10 @@ class BaseStore(driver.Store):
 
                         chunk_name = "%s-%05d" % (location.obj, chunk_id)
 
-                        with self.reader_class(image_file, checksum,
-                                               chunk_size, verifier) as reader:
+                        with self.reader_class(
+                                image_file, checksum, os_hash_value,
+                                chunk_size, verifier,
+                                backend_group=self.backend_group) as reader:
                             if reader.is_zero_size is True:
                                 LOG.debug('Not writing zero-length chunk.')
                                 break
@@ -983,11 +1030,12 @@ class BaseStore(driver.Store):
                     if image_size == 0:
                         image_size = combined_chunks_size
 
-                    # Now we write the object manifest and return the
-                    # manifest's etag...
+                    # Now we write the object manifest in X-Object-Manifest
+                    # header as defined for Dynamic Large Objects (DLO) Mode.
+                    # This request does not include ETag as PUT request has not
+                    # actual content that we need to verify.
                     manifest = "%s/%s-" % (location.container, location.obj)
-                    headers = {'ETag': hashlib.md5(b"").hexdigest(),
-                               'X-Object-Manifest': manifest}
+                    headers = {'X-Object-Manifest': manifest}
 
                     # The ETag returned for the manifest is actually the
                     # MD5 hash of the concatenated checksums of the strings
@@ -1004,12 +1052,19 @@ class BaseStore(driver.Store):
                 # image data. We *really* should consider NOT returning
                 # the location attribute from GET /images/<ID> and
                 # GET /images/details
-                if sutils.is_multiple_swift_store_accounts_enabled(self.conf):
+                if sutils.is_multiple_swift_store_accounts_enabled(
+                        self.conf, backend=self.backend_group):
                     include_creds = False
                 else:
                     include_creds = True
+
+                metadata = {}
+                if self.backend_group:
+                    metadata['store'] = u"%s" % self.backend_group
+
                 return (location.get_uri(credentials_included=include_creds),
-                        image_size, obj_etag, {})
+                        image_size, obj_etag, os_hash_value.hexdigest(),
+                        metadata)
             except swiftclient.ClientException as e:
                 if e.http_status == http_client.CONFLICT:
                     msg = _("Swift already has an image at this location")
@@ -1086,11 +1141,15 @@ class BaseStore(driver.Store):
         :param container: Name of container to create
         :param connection: Connection to swift service
         """
+        if self.backend_group:
+            store_conf = getattr(self.conf, self.backend_group)
+        else:
+            store_conf = self.conf.glance_store
         try:
             connection.head_container(container)
         except swiftclient.ClientException as e:
             if e.http_status == http_client.NOT_FOUND:
-                if self.conf.glance_store.swift_store_create_container_on_put:
+                if store_conf.swift_store_create_container_on_put:
                     try:
                         msg = (_LI("Creating swift container %(container)s") %
                                {'container': container})
@@ -1163,13 +1222,18 @@ class BaseStore(driver.Store):
         msg = _("There is no Connection Manager implemented for %s class.")
         raise NotImplementedError(msg % self.__class__.__name__)
 
+    def _set_url_prefix(self, context=None):
+        raise NotImplementedError()
+
 
 class SingleTenantStore(BaseStore):
     EXAMPLE_URL = "swift://<USER>:<KEY>@<AUTH_ADDRESS>/<CONTAINER>/<FILE>"
 
-    def __init__(self, conf):
-        super(SingleTenantStore, self).__init__(conf)
-        self.ref_params = sutils.SwiftParams(self.conf).params
+    def __init__(self, conf, backend=None):
+        super(SingleTenantStore, self).__init__(conf, backend=backend)
+        self.backend_group = backend
+        self.ref_params = sutils.SwiftParams(self.conf,
+                                             backend=backend).params
 
     def configure(self, re_raise_bsc=False):
         # set configuration before super so configure_add can override
@@ -1182,7 +1246,15 @@ class SingleTenantStore(BaseStore):
         super(SingleTenantStore, self).configure(re_raise_bsc=re_raise_bsc)
 
     def configure_add(self):
-        default_ref = self.conf.glance_store.default_swift_reference
+        if self.backend_group:
+            default_ref = getattr(self.conf,
+                                  self.backend_group).default_swift_reference
+            self.container = getattr(self.conf,
+                                     self.backend_group).swift_store_container
+        else:
+            default_ref = self.conf.glance_store.default_swift_reference
+            self.container = self.conf.glance_store.swift_store_container
+
         default_swift_reference = self.ref_params.get(default_ref)
         if default_swift_reference:
             self.auth_address = default_swift_reference.get('auth_address')
@@ -1195,7 +1267,7 @@ class SingleTenantStore(BaseStore):
             self.scheme = 'swift+http'
         else:
             self.scheme = 'swift+https'
-        self.container = self.conf.glance_store.swift_store_container
+
         self.auth_version = default_swift_reference.get('auth_version')
         self.user = default_swift_reference.get('user')
         self.key = default_swift_reference.get('key')
@@ -1211,6 +1283,50 @@ class SingleTenantStore(BaseStore):
             LOG.error(reason)
             raise exceptions.BadStoreConfiguration(store_name="swift",
                                                    reason=reason)
+        if self.backend_group:
+            self._set_url_prefix()
+
+    def _get_credstring(self):
+        if self.user and self.key:
+            return '%s:%s' % (urllib.parse.quote(self.user),
+                              urllib.parse.quote(self.key))
+        return ''
+
+    def _set_url_prefix(self, context=None):
+        auth_or_store_url = self.auth_address
+        if auth_or_store_url.startswith('http://'):
+            auth_or_store_url = auth_or_store_url[len('http://'):]
+        elif auth_or_store_url.startswith('https://'):
+            auth_or_store_url = auth_or_store_url[len('https://'):]
+
+        credstring = self._get_credstring()
+        auth_or_store_url = auth_or_store_url.strip('/')
+        container = self.container.strip('/')
+
+        if sutils.is_multiple_swift_store_accounts_enabled(
+                self.conf, backend=self.backend_group):
+            include_creds = False
+        else:
+            include_creds = True
+
+        if not include_creds:
+            store = getattr(self.conf,
+                            self.backend_group).default_swift_reference
+
+            self._url_prefix = '%s://%s/%s/' % (
+                'swift+config', store, container)
+            return
+
+        if self.scheme == 'swift+config':
+            if self.ssl_enabled:
+                self.scheme = 'swift+https'
+            else:
+                self.scheme = 'swift+http'
+        if credstring != '':
+            credstring = "%s@" % credstring
+
+        self._url_prefix = '%s://%s%s/%s/' % (
+            self.scheme, credstring, auth_or_store_url, container)
 
     def create_location(self, image_id, context=None):
         container_name = self.get_container_name(image_id, self.container)
@@ -1220,7 +1336,8 @@ class SingleTenantStore(BaseStore):
                  'auth_or_store_url': self.auth_address,
                  'user': self.user,
                  'key': self.key}
-        return StoreLocation(specs, self.conf)
+        return StoreLocation(specs, self.conf,
+                             backend_group=self.backend_group)
 
     def get_container_name(self, image_id, default_image_container):
         """
@@ -1238,8 +1355,14 @@ class SingleTenantStore(BaseStore):
         :param default_image_container: container name from
                ``swift_store_container``
         """
-        seed_num_chars = \
-            self.conf.glance_store.swift_store_multiple_containers_seed
+        if self.backend_group:
+            seed_num_chars = getattr(
+                self.conf,
+                self.backend_group).swift_store_multiple_containers_seed
+        else:
+            seed_num_chars = \
+                self.conf.glance_store.swift_store_multiple_containers_seed
+
         if seed_num_chars is None \
                 or seed_num_chars < 0 or seed_num_chars > 32:
             reason = _("An integer value between 0 and 32 is required for"
@@ -1331,7 +1454,7 @@ class SingleTenantStore(BaseStore):
             project_domain_id=self.project_domain_id,
             project_domain_name=self.project_domain_name)
 
-        sess = ks_session.Session(auth=password)
+        sess = ks_session.Session(auth=password, verify=not self.insecure)
         return ks_client.Client(session=sess)
 
     def get_manager(self, store_location, context=None, allow_reauth=False):
@@ -1345,7 +1468,12 @@ class MultiTenantStore(BaseStore):
     EXAMPLE_URL = "swift://<SWIFT_URL>/<CONTAINER>/<FILE>"
 
     def _get_endpoint(self, context):
-        self.container = self.conf.glance_store.swift_store_container
+        if self.backend_group:
+            self.container = getattr(self.conf,
+                                     self.backend_group).swift_store_container
+        else:
+            self.container = self.conf.glance_store.swift_store_container
+
         if context is None:
             reason = _("Multi-tenant Swift storage requires a context.")
             raise exceptions.BadStoreConfiguration(store_name="swift",
@@ -1418,7 +1546,13 @@ class MultiTenantStore(BaseStore):
                  'container': self.container + '_' + str(image_id),
                  'obj': str(image_id),
                  'auth_or_store_url': ep}
-        return StoreLocation(specs, self.conf)
+        return StoreLocation(specs, self.conf,
+                             backend_group=self.backend_group)
+
+    def _set_url_prefix(self, context=None):
+        ep = self._get_endpoint(context)
+        self._url_prefix = "%s://%s:%s_" % (
+            self.scheme, ep, self.container)
 
     def get_connection(self, location, context=None):
         return swiftclient.Connection(
@@ -1430,8 +1564,14 @@ class MultiTenantStore(BaseStore):
 
     def init_client(self, location, context=None):
         # read client parameters from config files
-        ref_params = sutils.SwiftParams(self.conf).params
-        default_ref = self.conf.glance_store.default_swift_reference
+        ref_params = sutils.SwiftParams(self.conf,
+                                        backend=self.backend_group).params
+        if self.backend_group:
+            default_ref = getattr(self.conf,
+                                  self.backend_group).default_swift_reference
+        else:
+            default_ref = self.conf.glance_store.default_swift_reference
+
         default_swift_reference = ref_params.get(default_ref)
         if not default_swift_reference:
             reason = _("default_swift_reference %s is "
@@ -1448,11 +1588,15 @@ class MultiTenantStore(BaseStore):
         project_domain_name = default_swift_reference.get(
             'project_domain_name')
 
+        if self.backend_group:
+            self._set_url_prefix(context=context)
+
         # create client for multitenant user(trustor)
         trustor_auth = ks_identity.V3Token(auth_url=auth_address,
                                            token=context.auth_token,
                                            project_id=context.tenant)
-        trustor_sess = ks_session.Session(auth=trustor_auth)
+        trustor_sess = ks_session.Session(auth=trustor_auth,
+                                          verify=not self.insecure)
         trustor_client = ks_client.Client(session=trustor_sess)
         auth_ref = trustor_client.session.auth.get_auth_ref(trustor_sess)
         roles = [t['name'] for t in auth_ref['roles']]
@@ -1468,7 +1612,8 @@ class MultiTenantStore(BaseStore):
             user_domain_name=user_domain_name,
             project_domain_id=project_domain_id,
             project_domain_name=project_domain_name)
-        trustee_sess = ks_session.Session(auth=password)
+        trustee_sess = ks_session.Session(auth=password,
+                                          verify=not self.insecure)
         trustee_client = ks_client.Client(session=trustee_sess)
 
         # request glance user id - we will use it as trustee user
@@ -1494,13 +1639,20 @@ class MultiTenantStore(BaseStore):
         )
         # now we can authenticate against KS
         # as trustee of user who provided token
-        client_sess = ks_session.Session(auth=client_password)
+        client_sess = ks_session.Session(auth=client_password,
+                                         verify=not self.insecure)
         return ks_client.Client(session=client_sess)
 
     def get_manager(self, store_location, context=None, allow_reauth=False):
         # if global toggle is turned off then do not allow re-authentication
         # with trusts
-        if not self.conf.glance_store.swift_store_use_trusts:
+        if self.backend_group:
+            use_trusts = getattr(self.conf,
+                                 self.backend_group).swift_store_use_trusts
+        else:
+            use_trusts = self.conf.glance_store.swift_store_use_trusts
+
+        if not use_trusts:
             allow_reauth = False
 
         return connection_manager.MultiTenantConnectionManager(self,
@@ -1510,11 +1662,14 @@ class MultiTenantStore(BaseStore):
 
 
 class ChunkReader(object):
-    def __init__(self, fd, checksum, total, verifier=None):
+    def __init__(self, fd, checksum, os_hash_value, total, verifier=None,
+                 backend_group=None):
         self.fd = fd
         self.checksum = checksum
+        self.os_hash_value = os_hash_value
         self.total = total
         self.verifier = verifier
+        self.backend_group = backend_group
         self.bytes_read = 0
         self.is_zero_size = False
         self.byteone = fd.read(1)
@@ -1532,9 +1687,23 @@ class ChunkReader(object):
         if i > left:
             i = left
 
-        result = self.do_read(i)
+        # Note(rosmaita): under some circumstances in py3, a zero-byte
+        # read results in a non-byte value that then causes a "unicode
+        # objects must be encoded before hashing" error when we do the
+        # hash computations below.  (At least that seems to be what's
+        # happening in testing.)  So just fake a zero-byte read and let
+        # the current execution path continue.
+        # See https://bugs.launchpad.net/glance-store/+bug/1805332
+        # TODO(rosmaita): find what in the execution path is returning
+        # a native string instead of bytes and fix it.
+        if i == 0:
+            result = b''
+        else:
+            result = self.do_read(i)
+
         self.bytes_read += len(result)
         self.checksum.update(result)
+        self.os_hash_value.update(result)
         if self.verifier:
             self.verifier.update(result)
         return result
