@@ -70,7 +70,6 @@ class OptsTestCase(base.StoreBaseTest):
         expected_opt_names = [
             'default_store',
             'stores',
-            'store_capabilities_update_min_interval',
             'cinder_api_insecure',
             'cinder_ca_certificates_file',
             'cinder_catalog_info',
@@ -85,6 +84,7 @@ class OptsTestCase(base.StoreBaseTest):
             'cinder_volume_type',
             'default_swift_reference',
             'https_insecure',
+            'filesystem_store_chunk_size',
             'filesystem_store_datadir',
             'filesystem_store_datadirs',
             'filesystem_store_file_perm',
@@ -122,6 +122,8 @@ class OptsTestCase(base.StoreBaseTest):
             'swift_store_ssl_compression',
             'swift_store_use_trusts',
             'swift_store_user',
+            'swift_buffer_on_upload',
+            'swift_upload_buffer_dir',
             'vmware_insecure',
             'vmware_ca_file',
             'vmware_api_retry_count',
